using System;
using System.Drawing;
using System.Windows.Forms;
using Dapfor.Net.Diagnostics;
using Dapfor.Net.Ui;

namespace Dapfor.Net.Grid.Benchmark.AddingRows
{
    public partial class AddingRowsControl : UserControl
    {
        static readonly HighPerformanceTimer hiPerfTimer = new HighPerformanceTimer();
        private bool stopping;

        public AddingRowsControl()
        {
            InitializeComponent();

            grid.Appearance.EvenRows.BackColor = Color.FromArgb(30, 128, 128, 128);
            grid.Appearance.OddRows.BackColor = Color.FromArgb(30, 255, 255, 255);

            //Initialize header and columns
            grid.Headers.Add(Header.FromDataType(typeof(DataObject)));
            foreach (Column column in grid.Headers[0])
            {
                column.Width = 80;
            }

            grid.PaintRow += delegate(object sender, PaintRowEventArgs e)
            {
                e.Parts &= e.Parts ^ PaintPart.RowSelectorImages;

                e.PaintAll();
                e.Handled = true;

                using(StringFormat sf = new StringFormat())
                {
                    sf.Alignment = StringAlignment.Far;
                    e.Graphics.DrawString(e.Row.VisibleIndex.ToString(), SystemFonts.CaptionFont, SystemBrushes.InfoText, e.RowSelectorBounds, sf);    
                }
            };

            grid.PaintRectangle += delegate(object sender, PaintRectangleEventArgs e)
            {
                if (grid.Nodes.Count == 0)
                {
                    e.PaintAll();
                    e.Handled = true;

                    //Draw text and logo if there are no rows
                    Point pt = new Point(e.Grid.Width - Properties.Resources.logo.Width, e.Grid.Height - Properties.Resources.logo.Height);
                    e.Graphics.DrawImage(Properties.Resources.logo, new Rectangle(pt, Properties.Resources.logo.Size));

                    Rectangle rc = grid.ClientRectangle;
                    rc.Width -= Properties.Resources.logo.Width;
                    rc.Offset(10, 30);

                    using (Font f = new Font("verdana", 13, GraphicsUnit.Pixel))
                    using (Brush b = new SolidBrush(Color.FromArgb(0x55, 0x66, 0x55)))
                    using (StringFormat sf = new StringFormat())
                    {
                        sf.Alignment = StringAlignment.Near;
                        sf.LineAlignment = StringAlignment.Near;
                        sf.FormatFlags &= StringFormatFlags.NoWrap ^ sf.FormatFlags;
                        e.Graphics.DrawString("When a programmer develops an application, sooner or later he has to deal with its performance. To ensure application performance he has to think over its architecture, pick necessary components and use them correctly. Otherwise, on late stages of development, when he has a lot of code, improving performance becomes a complex task requiring changes of working and debugged code.\r\n\r\nPerformance of data insertion into a grid:\r\n\r\nIt is most efficient to insert elements in the end of the grid (this is true for all grids). This method of adding data basically doesnt depend on the number of already inserted elements. Adding data to the beginning of grid or to a sorted grid heavily depends on the number of already inserted elements. This is especially true for insertion to the beginning of a grid, which should be avoided.", f, b, rc, sf);
                    }
                }
            };
        }

        public void StopTest()
        {
            stopping = true;

            //Remove sort
            grid.Headers[0].SortedColumns.Clear();
            
            //Clear rows
            grid.Rows.Clear();
        }


        private void OnBtnStartClick(object sender, EventArgs e)
        {
            panel.Enabled = false;

            
            //Remove sort
            grid.Headers[0].SortedColumns.Clear();

            //Remove already added rows
            if (grid.Rows.Count > 0)
            {
                Logger.Info("Prepare .Net Grid and remove all existing rows");         
                grid.Rows.Clear();    
            }
            
            int rowCount = 500000;
            int tranche = 10000;

            //Limit amount of adding data
            if(rbToEnd.Checked)
            {
                Logger.Info(string.Format("Performance measurement when rows are inserted to the end of grid"));
                rowCount = 500000;
                tranche = 10000;
            }
            else if(rbToBeginning.Checked)
            {
                Logger.Info(string.Format("Performance measurement when rows are inserted to the beginning of grid"));
                rowCount = 20000;
                tranche = 1000;
            }
            else if (rbToSortedGrid.Checked)
            {
                Logger.Info(string.Format("Performance measurement when rows are inserted to sorted grid"));
                rowCount = 100000;
                tranche = 10000;
                grid.Headers[0][1].SortDirection = SortDirection.Ascending;
                grid.Headers[0][3].SortDirection = SortDirection.Descending;
            }

            stopping = false;
            hiPerfTimer.Start();

            for (int i = 0; (i < rowCount) && !stopping; ++i)
            {
                
                //Adding to the end or to the sorted container
                Row row = grid.Rows.Add(new DataObject());

                //Adding to the beginning
                if (rbToBeginning.Checked)
                {
                    row.ChildIndex = 0;
                }

                if ((i + 1) % tranche == 0 && i > 0)
                {
                    hiPerfTimer.Stop();
                    

                    double duration = hiPerfTimer.Duration;

                    int rate = duration > 0 ? (int) (1000*tranche/duration) : 0;

                    Logger.Info(string.Format("Rows {0} - {1} added in {2:0} msec. Rate: {3} rows/sec", i + 1 - tranche, i, duration, rate));
                    Application.DoEvents();

                    hiPerfTimer.Start();
                }
            }

            hiPerfTimer.Stop();
            panel.Enabled = true;

            Logger.Info(string.Format("Measurement completed"));
        }
    }
}