using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Threading;
using System.Windows.Forms;
using Dapfor.Net.Diagnostics;
using Dapfor.Net.Ui;
using Timer=System.Windows.Forms.Timer;

namespace Dapfor.Net.Grid.Benchmark.DataUpdating
{
    public partial class DataUpdatingEventDrivenModelControl : UserControl
    {
        readonly Random random = new Random();
        private bool measuring;
        private DateTime measureStart;
        private int updatesCount;
        private readonly List<int> updateScenarios = new List<int>();
        PerformanceCounter cpuCounter = new PerformanceCounter();
        
        private DateTime lastStat = DateTime.UtcNow;
        private int lastStatCount;
        readonly List<EventPublisherDataObject> dataObjects = new List<EventPublisherDataObject>();


        
        private Timer timer;

        public DataUpdatingEventDrivenModelControl()
        {
            InitializeComponent();

            //Initialize performance counters
            cpuCounter.CategoryName = "Processor";
            cpuCounter.CounterName = "% Processor Time";
            cpuCounter.InstanceName = "_Total";

            grid.Highlighting.Fading = false;


            //Initialize header and columns
            grid.Headers.Add(Header.FromDataType(typeof(EventPublisherDataObject)));
            foreach (Column column in grid.Headers[0])
            {
                column.Width = 80;
            }

            grid.PaintRow += delegate(object sender, PaintRowEventArgs e)
            {
                e.Parts &= e.Parts ^ PaintPart.RowSelectorImages;

                e.PaintAll();
                e.Handled = true;

                using (StringFormat sf = new StringFormat())
                {
                    sf.Alignment = StringAlignment.Far;
                    e.Graphics.DrawString(e.Row.VisibleIndex.ToString(), SystemFonts.CaptionFont, SystemBrushes.InfoText, e.RowSelectorBounds, sf);
                }
            };
            
            grid.ProcessUpdate += OnUpdateCell;

            grid.PaintRectangle += delegate(object sender, PaintRectangleEventArgs e)
            {
                if (grid.Nodes.Count == 0)
                {
                    e.PaintAll();
                    e.Handled = true;

                    //Draw text and logo if there are no rows
                    Point pt = new Point(e.Grid.Width - Properties.Resources.logo.Width, e.Grid.Height - Properties.Resources.logo.Height);
                    e.Graphics.DrawImage(Properties.Resources.logo, new Rectangle(pt, Properties.Resources.logo.Size));

                    Rectangle rc = grid.ClientRectangle;
                    rc.Width -= Properties.Resources.logo.Width;
                    rc.Offset(10, 30);

                    using (Font f = new Font("verdana", 13, GraphicsUnit.Pixel))
                    using (Brush b = new SolidBrush(Color.FromArgb(0x55, 0x66, 0x55)))
                    using (StringFormat sf = new StringFormat())
                    {
                        sf.Alignment = StringAlignment.Near;
                        sf.LineAlignment = StringAlignment.Near;
                        sf.FormatFlags &= StringFormatFlags.NoWrap ^ sf.FormatFlags;
                        e.Graphics.DrawString("Event-driven model assumes that data implements INotifyPropertyChanged interface.\r\n\r\nAll operations in a grid are performed only in GUI thread. If data changes in another thread, the grid performs the required synchronization, usually synchronous data transfer to GUI thread. Event-driven model may feature asynchronous transfer of notifications when the calling thread is not blocked.\r\n\r\nIn general, grid characteristics in this mode are similar to non-event mode. The main distinction is the ability to synchronize data with asynchronous call that doesnt block the called thread. This synchronization method consumes more resources, however it is more thread-safe and helps to prevent thread deadlocks.", f, b, rc, sf);
                    }
                }
            };
        }

        

        public void StopTest()
        {
            grid.StopLoadBalancingProcess();
            updatesCount = 0;
            measuring = true;
            measureStart = DateTime.UtcNow;
            lastStatCount = 0;
            lastStat = DateTime.UtcNow;
            gbStatistics.Visible = false;
            grid.Rows.Clear();
            btnStart.Enabled = true;
            cbEnableSort.Enabled = true;
            cbFading.Enabled = true;
            cbNonGuiThread.Enabled = true;
            cbAsynchronous.Enabled = true;
        }


        void StartTest()
        {
            if(updateScenarios.Count > 0)
            {
                Application.DoEvents();

                grid.Highlighting.Fading = cbFading.Checked;

                int rowsCount = updateScenarios[0];
                updateScenarios.RemoveAt(0);

                grid.StopLoadBalancingProcess();

                grid.Headers[0].SortedColumns.Clear();


                //Populate the list of databjects
                dataObjects.Clear();
                for (int i = 0; i < rowsCount; ++i)
                {
                    dataObjects.Add(new EventPublisherDataObject());
                }

                //Populate the grid
                grid.Rows.Clear();
                foreach (EventPublisherDataObject dataObject in dataObjects)
                {
                    grid.Rows.Add(dataObject);
                }


                if(cbEnableSort.Checked)
                {
                    grid.Headers[0][1].SortDirection = SortDirection.Ascending;
                    grid.Headers[0][3].SortDirection = SortDirection.Descending;
                }

                measuring = false;
                btnStart.Enabled = false;
                cbEnableSort.Enabled = false;
                cbFading.Enabled = false;
                cbNonGuiThread.Enabled = false;
                cbAsynchronous.Enabled = false;

                grid.StartLoadBalancingProcess();
            }
            else
            {
                measuring = false;
                btnStart.Enabled = true;
                cbEnableSort.Enabled = true;
                cbFading.Enabled = true;
                cbNonGuiThread.Enabled = true;
                cbAsynchronous.Enabled = true;

                grid.StopLoadBalancingProcess();

                Logger.Info(string.Format("Measurement completed"));
            }
        }



        private void OnUpdateCell(object sender, EventArgs e)
        {
            if(grid.IsLimitFound && !measuring)
            {
                updatesCount = 0;
                measuring = true;
                measureStart = DateTime.UtcNow;
                lastStatCount = 0;
                lastStat = DateTime.UtcNow;
            }


            //Updates random cell in the grid;
            EventPublisherDataObject dataObject = dataObjects.Count > 0 ? dataObjects[random.Next() % dataObjects.Count] : null;
            int field = random.Next()%grid.Headers[0].Count;

            if (dataObject != null)
            {
                int value = random.Next();
                if(cbNonGuiThread.Checked)
                {
                    ThreadPool.QueueUserWorkItem(delegate
                    {
                        UpdateDataobject(dataObject, field, value);
                    });    
                }
                else
                {
                    UpdateDataobject(dataObject, field, value);
                }
            }

            updatesCount++;
            if(measuring)
            {
                //Measure refresh rate for 5 seconds
                if(DateTime.UtcNow - measureStart > TimeSpan.FromSeconds(5))
                {
                    grid.StopLoadBalancingProcess();

                    int visibleRows = grid.Rows.BottomVisibleRow != null && grid.Rows.TopVisibleRow != null
                        ? (grid.Rows.BottomVisibleRow.VisibleIndex - grid.Rows.TopVisibleRow.VisibleIndex)
                        : 0;
                    long visibleCells = grid.Headers[0].Count * visibleRows;
                    long totalCells = grid.Headers[0].Count * grid.Rows.Count;

                    double rate = totalCells != 0 ? (double)visibleCells / totalCells : 0;


                    Logger.Info(string.Format("Grid of {0} rows, size = {2}x{3}. Maximum refresh rate: {1} updates/sec, in visible bounds: {4:0} updates/sec", 
                                              grid.Rows.Count, 
                                              updatesCount / 5, 
                                              grid.ClientSize.Width,
                                              grid.ClientSize.Height,
                                              rate * updatesCount / 5));

                    //Execute the next scenario
                    StartTest();
                }
            }

            DisplayStatistics();
        }

        private static void UpdateDataobject(EventPublisherDataObject dataObject, int field, int value)
        {
            if(dataObject != null)
            {
                switch (field)
                {
                    case 0:
                        dataObject.Value0 = value;
                        break;
                    case 1:
                        dataObject.Value1 = value;
                        break;
                    case 2:
                        dataObject.Value2 = value;
                        break;
                    case 3:
                        dataObject.Value3 = value;
                        break;
                    case 4:
                        dataObject.Value4 = value;
                        break;
                    case 5:
                        dataObject.Value5 = value;
                        break;
                    case 6:
                        dataObject.Value6 = value;
                        break;
                    case 7:
                        dataObject.Value7 = value;
                        break;
                    case 8:
                        dataObject.Value8 = value;
                        break;
                    case 9:
                        dataObject.Value9 = value;
                        break;
                }
            }
        }

        private void DisplayStatistics()
        {
            if(DateTime.UtcNow - lastStat > TimeSpan.FromMilliseconds(500))
            {
                lbCPULoadRate.Text = string.Format("CPU load rate: {0:0} %", cpuCounter.NextValue());
                lbGridSize.Text = string.Format("Grid size: {0}x{1}", grid.ClientSize.Width, grid.ClientSize.Height);
                double updates = 1000 * (updatesCount - lastStatCount) / (DateTime.UtcNow - lastStat).TotalMilliseconds;
                lbUpdateRate.Text = string.Format("Update rate: {0:0} updates/sec", updates);

                int visibleRows = grid.Rows.BottomVisibleRow != null && grid.Rows.TopVisibleRow != null
                        ? (grid.Rows.BottomVisibleRow.VisibleIndex - grid.Rows.TopVisibleRow.VisibleIndex)
                        : 0;
                long visibleCells = grid.Headers[0].Count * visibleRows;
                long totalCells = grid.Headers[0].Count * grid.Rows.Count;
                
                double rate = totalCells != 0 ? (double)visibleCells / totalCells : 0;
                lbVisibleBoundsUpdateRate.Text = string.Format("Update rate in visible bounds: {0:0} updates/sec", rate * updates);

                gbStatistics.Visible = true;

                lastStatCount = updatesCount;
                lastStat = DateTime.UtcNow;
            }
        }

        private void btnStart_Click(object sender, EventArgs e)
        {
            updateScenarios.Add(10);
            updateScenarios.Add(20);
            updateScenarios.Add(50);
            updateScenarios.Add(100);
            updateScenarios.Add(200);
            updateScenarios.Add(500);
            updateScenarios.Add(1000);
            updateScenarios.Add(5000);
            updateScenarios.Add(10000);
            updateScenarios.Add(20000);
            updateScenarios.Add(50000);
            updateScenarios.Add(100000);

            Logger.Info(string.Format("Application measures performance of .Net Grid. It may take a little time..."));

            StartTest();
        }

        private void cbAsynchronous_CheckedChanged(object sender, EventArgs e)
        {
            grid.Threadsafety.AsynchronousUpdates = cbAsynchronous.Checked;
        }

        private void cbNonGuiThread_CheckedChanged(object sender, EventArgs e)
        {

        }
    }
}