using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Globalization;
using System.Threading;
using System.Windows.Forms;
using Dapfor.Net.Diagnostics;
using Dapfor.Net.Ui;
using Timer=System.Windows.Forms.Timer;

namespace Dapfor.Net.Grid.Benchmark.Filtering
{
    public partial class DataFilteringControl : UserControl
    {
        readonly Random random = new Random();
        private bool measuring;
        private DateTime measureStart;
        private int updatesCount;
        private readonly List<int> updateScenarios = new List<int>();
        PerformanceCounter cpuCounter = new PerformanceCounter();
        
        private DateTime lastStat = DateTime.UtcNow;
        private int lastStatCount;
        private int minValue;
        private int maxValue;

        
        private Timer timer;

        public DataFilteringControl()
        {
            InitializeComponent();

            //Initialize performance counters
            cpuCounter.CategoryName = "Processor";
            cpuCounter.CounterName = "% Processor Time";
            cpuCounter.InstanceName = "_Total";

            grid.Highlighting.Fading = false;


            //Initialize header and columns
            grid.Headers.Add(Header.FromDataType(typeof (DataObject)));
            foreach (Column column in grid.Headers[0])
            {
                column.Width = 80;
            }
            
            grid.ProcessUpdate += OnUpdateCell;

            OnFilterChanged(this, EventArgs.Empty);
            grid.Filter = new Filter(OnFilterRow);

            grid.PaintRectangle += delegate(object sender, PaintRectangleEventArgs e)
            {
                if (grid.Nodes.Count == 0)
                {
                    e.PaintAll();
                    e.Handled = true;

                    //Draw text and logo if there are no rows
                    Point pt = new Point(e.Grid.Width - Properties.Resources.logo.Width, e.Grid.Height - Properties.Resources.logo.Height);
                    e.Graphics.DrawImage(Properties.Resources.logo, new Rectangle(pt, Properties.Resources.logo.Size));

                    Rectangle rc = grid.ClientRectangle;
                    rc.Width -= Properties.Resources.logo.Width;
                    rc.Offset(10, 30);

                    using (Font f = new Font("verdana", 13, GraphicsUnit.Pixel))
                    using (Brush b = new SolidBrush(Color.FromArgb(0x55, 0x66, 0x55)))
                    using (StringFormat sf = new StringFormat())
                    {
                        sf.Alignment = StringAlignment.Near;
                        sf.LineAlignment = StringAlignment.Near;
                        sf.FormatFlags &= StringFormatFlags.NoWrap ^ sf.FormatFlags;
                        e.Graphics.DrawString("Filtering manages row visibility. A developer may set his own filters to manage filtering. Every time when new data is added to a grid or when data is modified, the grid checks whether this data meets filtering conditions.\r\n\r\nPerformance heavily depends on the number of rows and on vertical scrollbar position. When the thumb is in the middle, grid has to repaint more rows because when visibility status of a higher-standing row changes, all the following rows change their positions. ", f, b, rc, sf);
                    }
                }
            };
        }

        public void StopTest()
        {
            grid.StopLoadBalancingProcess();
            updatesCount = 0;
            measuring = true;
            measureStart = DateTime.UtcNow;
            lastStatCount = 0;
            lastStat = DateTime.UtcNow;
            gbStatistics.Visible = false;
            grid.Rows.Clear();
            btnStart.Enabled = true;
        }


        void StartTest()
        {
            if(updateScenarios.Count > 0)
            {
                int rowsCount = updateScenarios[0];
                updateScenarios.RemoveAt(0);

                grid.StopLoadBalancingProcess();

                grid.Headers[0].SortedColumns.Clear();


                //Populate the grid
                grid.Rows.Clear();
                for (int i = 0; i < rowsCount; ++i)
                {
                    DataObject dataObject = new DataObject();
                    dataObject.Value0 = random.Next((int)(0.9 * minValue), (int)(1.1 * maxValue));
                    dataObject.Value1 = random.Next((int)(0.9 * minValue), (int)(1.1 * maxValue));
                    dataObject.Value2 = random.Next((int)(0.9 * minValue), (int)(1.1 * maxValue));
                    dataObject.Value3 = random.Next((int)(0.9 * minValue), (int)(1.1 * maxValue));
                    dataObject.Value4 = random.Next((int)(0.9 * minValue), (int)(1.1 * maxValue));
                    dataObject.Value5 = random.Next((int)(0.9 * minValue), (int)(1.1 * maxValue));
                    dataObject.Value6 = random.Next((int)(0.9 * minValue), (int)(1.1 * maxValue));
                    dataObject.Value7 = random.Next((int)(0.9 * minValue), (int)(1.1 * maxValue));
                    dataObject.Value8 = random.Next((int)(0.9 * minValue), (int)(1.1 * maxValue));
                    dataObject.Value9 = random.Next((int)(0.9 * minValue), (int)(1.1 * maxValue));

                    grid.Rows.Add(dataObject);
                }

                measuring = false;
                btnStart.Enabled = false;
                grid.StartLoadBalancingProcess();
            }
            else
            {
                measuring = false;
                btnStart.Enabled = true;
                grid.StopLoadBalancingProcess();

                Logger.Info(string.Format("Measurement completed"));
            }
        }



        private void OnUpdateCell(object sender, EventArgs e)
        {
            if(grid.IsLimitFound && !measuring)
            {
                updatesCount = 0;
                measuring = true;
                measureStart = DateTime.UtcNow;
                lastStatCount = 0;
                lastStat = DateTime.UtcNow;
            }


            //Updates random cell in the grid;
            Row row = grid.Nodes.Count > 0 ? grid.Nodes[random.Next() % grid.Nodes.Count] : null;
            Cell cell = row != null ? row[random.Next() % grid.Headers[0].Count] : null;

            if (cell != null)
            {
                int value = random.Next((int) (0.9 * minValue),  (int) (1.1 * maxValue) );
                if(cbNonGuiThread.Checked)
                {
                    ThreadPool.QueueUserWorkItem(delegate
                    {
                        cell.Value = value;
                    });    
                }
                else
                {
                    cell.Value = value;
                }
            }

            updatesCount++;
            if(measuring)
            {
                //Measure refresh rate for 5 seconds
                if(DateTime.UtcNow - measureStart > TimeSpan.FromSeconds(5))
                {
                    grid.StopLoadBalancingProcess();

                    int visibleRows = grid.Rows.BottomVisibleRow != null && grid.Rows.TopVisibleRow != null
                        ? (grid.Rows.BottomVisibleRow.VisibleIndex - grid.Rows.TopVisibleRow.VisibleIndex)
                        : 0;

                    long visibleCells = grid.Headers[0].Count * visibleRows;
                    long totalCells = grid.Headers[0].Count * grid.Nodes.Count;

                    double rate = totalCells != 0 ? (double)visibleCells / totalCells : 0;


                    Logger.Info(string.Format("Grid of {0} rows, size = {2}x{3}. Maximum filtering rate: {1} updates/sec, in visible bounds: {4:0} updates/sec", 
                                              grid.Nodes.Count, 
                                              updatesCount / 5, 
                                              grid.ClientSize.Width,
                                              grid.ClientSize.Height,
                                              rate * updatesCount / 5));

                    //Execute the next scenario
                    StartTest();
                }
            }

            DisplayStatistics();
        }

        private void DisplayStatistics()
        {
            if(DateTime.UtcNow - lastStat > TimeSpan.FromMilliseconds(500))
            {
                lbCPULoadRate.Text = string.Format("CPU load rate: {0:0} %", cpuCounter.NextValue());
                lbGridSize.Text = string.Format("Grid size: {0}x{1}", grid.ClientSize.Width, grid.ClientSize.Height);
                double updates = 1000 * (updatesCount - lastStatCount) / (DateTime.UtcNow - lastStat).TotalMilliseconds;
                lbUpdateRate.Text = string.Format("Update rate: {0:0} updates/sec", updates);

                int visibleRows = grid.Rows.BottomVisibleRow != null && grid.Rows.TopVisibleRow != null 
                    ? (grid.Rows.BottomVisibleRow.VisibleIndex - grid.Rows.TopVisibleRow.VisibleIndex) 
                    : 0;

                long visibleCells = grid.Headers[0].Count * visibleRows;
                long totalCells = grid.Headers[0].Count * grid.Nodes.Count;
                
                double rate = totalCells != 0 ? (double)visibleCells / totalCells : 0;
                lbVisibleBoundsUpdateRate.Text = string.Format("Update rate in visible bounds: {0:0} updates/sec", rate * updates);

                gbStatistics.Visible = true;

                lastStatCount = updatesCount;
                lastStat = DateTime.UtcNow;
            }
        }

        private void btnStart_Click(object sender, EventArgs e)
        {
            updateScenarios.Add(10);
            updateScenarios.Add(20);
            updateScenarios.Add(50);
            updateScenarios.Add(100);
            updateScenarios.Add(200);
            updateScenarios.Add(500);
            updateScenarios.Add(1000);
            updateScenarios.Add(5000);
            updateScenarios.Add(10000);
            updateScenarios.Add(20000);
            updateScenarios.Add(50000);
            updateScenarios.Add(100000);

            Logger.Info(string.Format("Application measures performance of .Net Grid. It may take a little time..."));

            StartTest();
        }

        private void OnFilterChanged(object sender, EventArgs e)
        {
            if(!string.IsNullOrEmpty(tbMinValue.Text))
            {
                int.TryParse(tbMinValue.Text, NumberStyles.Number, CultureInfo.InvariantCulture, out minValue);
            }
            if(!string.IsNullOrEmpty(tbMaxValue.Text))
            {
                int.TryParse(tbMaxValue.Text, NumberStyles.Number, CultureInfo.InvariantCulture, out maxValue);
            }
            grid.FilterRefresh();
        }

        private bool OnFilterRow(Row row)
        {
            bool filtered = false;
            foreach (Column column in grid.Headers[0])
            {
                Cell cell = row[column.Id];
                if(cell != null)
                {
                    int value = (int) cell.Value;
                    filtered = value < minValue || value > maxValue;
                    if(filtered)
                    {
                        break;
                    }
                }
            }
            return filtered;
        }
    }
}